@@ -9,6 +9,12 @@ from django.contrib.staticfiles.storage import (
     ManifestStaticFilesStorage,
 )
 from django.core.files.storage import FileSystemStorage
+from django.core.exceptions import ImproperlyConfigured
+
+from boto.s3.connection import (
+    SubdomainCallingFormat,
+    OrdinaryCallingFormat,
+)
 
 from six.moves.urllib import parse
 from storages.backends import s3boto
@@ -26,16 +32,24 @@ parse.uses_netloc.append('djfs')
 
 class S3MediaStorage(s3boto.S3BotoStorage):
     def __init__(self):
+        bucket_name = settings.AWS_MEDIA_STORAGE_BUCKET_NAME
+
+        if '.' in bucket_name:
+            calling_format = OrdinaryCallingFormat()
+        else:
+            calling_format = SubdomainCallingFormat()
+
         # We cannot use a function call or a partial here. Instead, we have to
         # create a subclass because django tries to recreate a new object by
         # calling the __init__ of the returned object (with no arguments).
         super(S3MediaStorage, self).__init__(
             access_key=settings.AWS_MEDIA_ACCESS_KEY_ID,
             secret_key=settings.AWS_MEDIA_SECRET_ACCESS_KEY,
-            bucket_name=settings.AWS_MEDIA_STORAGE_BUCKET_NAME,
+            bucket_name=bucket_name,
             location=settings.AWS_MEDIA_BUCKET_PREFIX,
             host=settings.AWS_MEDIA_STORAGE_HOST,
             custom_domain=settings.AWS_MEDIA_DOMAIN,
+            calling_format=calling_format,
             # Setting an ACL requires us to grant the user the PutObjectAcl
             # permission as well, even if it matches the default bucket ACL.
             # XXX: Ideally we would thus set it to `None`, but due to how
@@ -99,8 +113,12 @@ class S3MediaStorage(s3boto.S3BotoStorage):
             total += 1
 
             if new_headers != old_headers:
-                key.copy(self.bucket.name, key,
-                         metadata=new_headers, preserve_acl=True)
+                key.copy(
+                    self.bucket.name, key,
+                    metadata=new_headers,
+                    preserve_acl=True,
+                    validate_dst_bucket=False,
+                )
                 updated += 1
 
         return updated, total
@@ -115,15 +133,26 @@ def parse_storage_url(url):
     config['DEFAULT_FILE_STORAGE'] = SCHEMES[scheme[0]]
 
     if scheme[0] == 's3':
-        os.environ['S3_USE_SIGV4'] = 'True'
+        query = parse.parse_qs(url.query)
+        media_domain = query.get('domain', [None])[0]
+        signature_ver = query.get('auth', ['s3v4'])[0]
+        endpoint = url.hostname.rsplit('.', 3)
+        bucket_name = endpoint[0]
+        storage_host = '.'.join(endpoint[1:])
 
-        media_domain = parse.parse_qs(url.query).get('domain', [None])[0]
+        if signature_ver == 's3v4':
+            os.environ['S3_USE_SIGV4'] = 'True'
+        elif signature_ver == 's3':
+            os.environ['S3_USE_SIGV4'] = ''
+        else:
+            raise ImproperlyConfigured('Unknown signature version: {}'
+                                       .format(signature_ver))
 
         config.update({
             'AWS_MEDIA_ACCESS_KEY_ID': parse.unquote(url.username or ''),
             'AWS_MEDIA_SECRET_ACCESS_KEY': parse.unquote(url.password or ''),
-            'AWS_MEDIA_STORAGE_BUCKET_NAME': url.hostname.split('.', 1)[0],
-            'AWS_MEDIA_STORAGE_HOST': url.hostname.split('.', 1)[1],
+            'AWS_MEDIA_STORAGE_BUCKET_NAME': bucket_name,
+            'AWS_MEDIA_STORAGE_HOST': storage_host,
             'AWS_MEDIA_BUCKET_PREFIX': url.path.lstrip('/'),
             'AWS_MEDIA_DOMAIN': media_domain,
         })
@@ -168,7 +197,7 @@ def parse_storage_url(url):
     return config
 
 
-class GzippedStaticFilesMixin(object):
+class GZippedStaticFilesMixin(object):
     """
     Static files storage mixin to create a gzipped version of each static
     file, so that web servers (e.g. uWSGI) can take advantage of it and
@@ -186,7 +215,7 @@ class GzippedStaticFilesMixin(object):
     def gzip_path(self, path):
         gz_path = path + '.gz'
         with self.open(path) as f_in:
-            with self.open(gz_path, 'w') as f_out:
+            with self.open(gz_path, 'wb') as f_out:
                 with gzip.GzipFile(fileobj=f_out) as gz_out:
                     shutil.copyfileobj(f_in, gz_out)
         return gz_path
@@ -200,7 +229,7 @@ class GzippedStaticFilesMixin(object):
             yield os.path.join(path, file)
 
     def post_process(self, paths, dry_run=False, **options):
-        superclass = super(GzippedStaticFilesMixin, self)
+        superclass = super(GZippedStaticFilesMixin, self)
         if hasattr(superclass, 'post_process'):
             post_processed = (
                 superclass.post_process(paths, dry_run=dry_run, **options)
@@ -222,11 +251,11 @@ class GzippedStaticFilesMixin(object):
                 self.gzip_path(path)
 
 
-class GzippedStaticFilesStorage(GzippedStaticFilesMixin,
+class GZippedStaticFilesStorage(GZippedStaticFilesMixin,
                                 StaticFilesStorage):
     pass
 
 
-class ManifestGzippedStaticFilesStorage(GzippedStaticFilesMixin,
+class ManifestGZippedStaticFilesStorage(GZippedStaticFilesMixin,
                                         ManifestStaticFilesStorage):
     pass
